@@ -1,5 +1,6 @@
 # -*- Mode: Python; py-indent-offset: 4 -*-
 import sys
+from copy import *
 
 def get_valid_scheme_definitions(defs):
     return [x for x in defs if isinstance(x, tuple) and len(x) >= 2]
@@ -11,11 +12,17 @@ class Parameter(object):
         self.pname = pname
         self.pdflt = pdflt
         self.pnull = pnull
-        self.prop = prop
+        
     def __len__(self): return 4
     def __getitem__(self, i):
         return (self.ptype, self.pname, self.pdflt, self.pnull)[i]
 
+    def merge(self, old):
+        if old.pdflt is not None:
+            self.pdflt = old.pdflt
+        if old.pnull is not None:
+            self.pnull = old.pnull
+
 # Parameter for property based constructors
 class Property(object):
     def __init__(self, pname, optional, argname):
@@ -23,6 +30,12 @@ class Property(object):
         self.optional = optional
         self.argname = argname
 
+    def merge(self, old):
+        if old.optional is not None:
+            self.optional = old.optional
+        if old.argname is not None:
+            self.argname = old.argname
+
 
 class Definition:
     def __init__(self, *args):
@@ -55,6 +68,7 @@ class ObjectDef(Definition):
         self.typecode = None
 	self.fields = []
         self.implements = []
+        self.class_init_func = None
 	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
@@ -100,7 +114,9 @@ class InterfaceDef(Definition):
 	self.module = None
 	self.c_name = None
         self.typecode = None
+        self.vtable = None
 	self.fields = []
+        self.interface_info = None
 	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
@@ -108,6 +124,10 @@ class InterfaceDef(Definition):
 		self.c_name = arg[1]
 	    elif arg[0] == 'gtype-id':
 		self.typecode = arg[1]
+	    elif arg[0] == 'vtable':
+		self.vtable = arg[1]
+        if self.vtable is None:
+            self.vtable = self.c_name + "Iface"
     def write_defs(self, fp=sys.stdout):
 	fp.write('(define-interface ' + self.name + '\n')
 	if self.module:
@@ -240,7 +260,7 @@ class PointerDef(Definition):
             fp.write('  )\n')
 	fp.write(')\n\n')
 
-class MethodDef(Definition):
+class MethodDefBase(Definition):
     def __init__(self, name, *args):
         dump = 0
 	self.name = name
@@ -289,27 +309,29 @@ class MethodDef(Definition):
 
         if self.caller_owns_return is None and self.ret is not None:
             self.guess_return_value_ownership()
-        for item in ('c_name', 'of_object'):
-            if self.__dict__[item] == None:
-                self.write_defs(sys.stderr)
-                raise RuntimeError, "definition missing required %s" % (item,)
             
-    def merge(self, old):
+    def merge(self, old, parmerge):
+        self.caller_owns_return = old.caller_owns_return
+        self.varargs = old.varargs
 	# here we merge extra parameter flags accross to the new object.
+        if not parmerge:
+            self.params = deepcopy(old.params)
+            return
 	for i in range(len(self.params)):
 	    ptype, pname, pdflt, pnull = self.params[i]
 	    for p2 in old.params:
 		if p2[1] == pname:
 		    self.params[i] = (ptype, pname, p2[2], p2[3])
 		    break
-    def write_defs(self, fp=sys.stdout):
-	fp.write('(define-method ' + self.name + '\n')
+    def _write_defs(self, fp=sys.stdout):
 	if self.of_object != (None, None):
 	    fp.write('  (of-object "' + self.of_object + '")\n')
 	if self.c_name:
 	    fp.write('  (c-name "' + self.c_name + '")\n')
 	if self.typecode:
 	    fp.write('  (gtype-id "' + self.typecode + '")\n')
+        if self.caller_owns_return:
+	    fp.write('  (caller-owns-return #t)\n')
 	if self.ret:
 	    fp.write('  (return-type "' + self.ret + '")\n')
 	if self.deprecated:
@@ -322,8 +344,28 @@ class MethodDef(Definition):
                 if pnull: fp.write(' (null-ok)')
                 fp.write(')\n')
             fp.write('  )\n')
+	if self.varargs:
+	    fp.write('  (varargs #t)\n')
 	fp.write(')\n\n')
 
+
+class MethodDef(MethodDefBase):
+    def __init__(self, name, *args):
+        MethodDefBase.__init__(self, name, *args)
+        for item in ('c_name', 'of_object'):
+            if self.__dict__[item] == None:
+                self.write_defs(sys.stderr)
+                raise RuntimeError, "definition missing required %s" % (item,)
+        
+    def write_defs(self, fp=sys.stdout):
+	fp.write('(define-method ' + self.name + '\n')
+        self._write_defs(fp)
+
+class VirtualDef(MethodDefBase):
+    def write_defs(self, fp=sys.stdout):
+	fp.write('(define-virtual ' + self.name + '\n')
+        self._write_defs(fp)
+
 class FunctionDef(Definition):
     def __init__(self, name, *args):
         dump = 0
@@ -396,14 +438,33 @@ class FunctionDef(Definition):
 
     _method_write_defs = MethodDef.__dict__['write_defs']
 
-    def merge(self, old):
+    def merge(self, old, parmerge):
+        self.caller_owns_return = old.caller_owns_return
+        self.varargs = old.varargs
+        if not parmerge:
+            self.params = deepcopy(old.params)
+            return
 	# here we merge extra parameter flags accross to the new object.
-	for i in range(len(self.params)):
-	    ptype, pname, pdflt, pnull = self.params[i]
-	    for p2 in old.params:
-		if p2[1] == pname:
-		    self.params[i] = (ptype, pname, p2[2], p2[3])
-		    break
+        def merge_param(param):
+	    for old_param in old.params:
+		if old_param.pname == param.pname:
+                    if isinstance(old_param, Property):
+                        # h2def never scans Property's, therefore if
+                        # we have one it was manually written, so we
+                        # keep it.
+                        return deepcopy(old_param)
+                    else:
+                        param.merge(old_param)
+                        return param
+            raise RuntimeError, "could not find %s in old_parameters %r" % (
+                param.pname, [p.pname for p in old.params])
+        try:
+            self.params = map(merge_param, self.params)
+        except RuntimeError:
+            # parameter names changed and we can't find a match; it's
+            # safer to keep the old parameter list untouched.
+            self.params = deepcopy(old.params)
+        
 	if not self.is_constructor_of:
             try:
                 self.is_constructor_of = old.is_constructor_of
@@ -425,6 +486,8 @@ class FunctionDef(Definition):
 	    fp.write('  (c-name "' + self.c_name + '")\n')
 	if self.typecode:
 	    fp.write('  (gtype-id "' + self.typecode + '")\n')
+        if self.caller_owns_return:
+	    fp.write('  (caller-owns-return #t)\n')
 	if self.ret:
 	    fp.write('  (return-type "' + self.ret + '")\n')
 	if self.deprecated:
@@ -446,5 +509,8 @@ class FunctionDef(Definition):
                     fp.write(')\n')
                 fp.write('  )\n')
             else:
-                assert False, "strange parameter list"
+                assert False, "strange parameter list %r" % self.params[0]
+	if self.varargs:
+	    fp.write('  (varargs #t)\n')
+
 	fp.write(')\n\n')
